package org.globus.transfer.reliable.service.test.connection;

import junit.framework.TestCase;
import commonj.timers.TimerManager;
import org.globus.rft.generated.RFTOptionsType;
import org.ietf.jgss.GSSCredential;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.gssapi.GlobusGSSCredentialImpl;
import org.globus.transfer.reliable.service.test.TestConfig;
import org.globus.transfer.reliable.service.cache.Connection;
import org.globus.transfer.reliable.service.cache.SingleConnectionImpl;
import org.globus.transfer.reliable.service.cache.MlsdConnectionImpl;
import org.globus.transfer.reliable.service.cache.ThirdPartyConnectionImpl;
import org.globus.transfer.reliable.service.cache.ConnectionCacheConfiguration;
import org.globus.transfer.reliable.service.cache.ConnectionManager;
import org.globus.transfer.reliable.service.cache.ConnectionParameter;
import org.globus.wsrf.impl.timer.TimerManagerImpl;

public class ConnectionManagerTest extends TestCase {

    private TimerManager timerManager;
    private ConnectionCacheConfiguration cacheConfig;
    private ConnectionManager connManager;
    public GSSCredential credential = null;
    RFTOptionsType rftOptions1;
    RFTOptionsType rftOptions2;
    private ConnectionParameter singleParams;
    private ConnectionParameter mlsdParams;
    private ConnectionParameter tpParams1;
    private ConnectionParameter tpParams2;
    private int timeout = 5000;
    
    public ConnectionManagerTest(String name) {
        super(name);
    }

    protected void setUp() throws Exception {
        
        super.setUp();

        this.timerManager = new TimerManagerImpl();
        this.cacheConfig = new ConnectionCacheConfiguration();
        this.timeout = 5000;
        cacheConfig.setConnectionTimeout(timeout);
        cacheConfig.setMaxNumberCachedFDs(10);
        cacheConfig.setSweeperInterval(1000);

        // create a connection manager
        this.connManager =  ConnectionManager.createNewInstance(
            cacheConfig, timerManager);

        this.credential = new GlobusGSSCredentialImpl(
            GlobusCredential.getDefaultCredential(),
            GSSCredential.INITIATE_AND_ACCEPT);

        // create different RFTOptionsType objects
        rftOptions1 = new RFTOptionsType();
        rftOptions2 = new RFTOptionsType();
        rftOptions1.setDcau(new Boolean(false));
        rftOptions2.setDcau(new Boolean(true));

        // create different connection paramater objects
        singleParams = new ConnectionParameter(
            TestConfig.sourceHost, TestConfig.sourcePort,
            TestConfig.sourcePath, TestConfig.subject,
            rftOptions1);
        mlsdParams = new ConnectionParameter(
            TestConfig.sourceHost, TestConfig.sourcePort,
            TestConfig.sourcePath, TestConfig.subject,
            rftOptions2);
        tpParams1 = new ConnectionParameter(
            TestConfig.sourceHost, TestConfig.sourcePort,
            TestConfig.sourcePath, TestConfig.destinationHost,
            TestConfig.destPort, TestConfig.destPath, TestConfig.subject,
            rftOptions1);
        tpParams2 = new ConnectionParameter(
            TestConfig.sourceHost, TestConfig.sourcePort,
            TestConfig.sourcePath, TestConfig.destinationHost,
            TestConfig.destPort, TestConfig.destPath, TestConfig.subject,
            rftOptions2);

        TestConfig.printConfig();
    }

    protected void tearDown() {
        
        this.timerManager.stop();
        this.connManager.closeAllConnections();
    }

    /**
     * Try to create a ConnectionManager with various invalid arguments.
     */
    public void testCreateNewInstance() {
      
        try {
            ConnectionManager.createNewInstance(null, this.timerManager);
            fail("null as argument of constructor of " +
                "ConnectionManager must fail");
        } catch (Exception e) {}

        try {
            ConnectionManager.createNewInstance(this.cacheConfig, null);
            fail("null as argument of constructor of " +
                "ConnectionManager must fail");
        } catch (Exception e) {}
    }
    
    public void testGetSingleConnectionFromEmptyCache() {
        
        assertTrue(this.connManager.getNumberCachedFDs() == 0);
        try {
            Connection conn = this.connManager.getConnection(
                ConnectionManager.SINGLE_CONNECTION, singleParams,
                credential, TestConfig.sourceSubject);
            assertTrue(this.connManager.getNumberCachedFDs() == 0);
            assertTrue(conn instanceof SingleConnectionImpl);
        } catch(Exception e) {
            fail("testGetSingleConnectionFromEmptyCache() failed: " + e);
        }
    }
    
    public void testGetMlsdConnectionFromEmptyCache() {
        
        assertTrue(this.connManager.getNumberCachedFDs() == 0);
        try {
            Connection conn = this.connManager.getConnection(
                ConnectionManager.MLSD_CONNECTION, mlsdParams,
                credential, TestConfig.sourceSubject);
            assertTrue(this.connManager.getNumberCachedFDs() == 0);
            assertTrue(conn instanceof MlsdConnectionImpl);
        } catch(Exception e) {
            fail("testGetMlsdConnectionFromEmptyCache() failed: " + e);
        }
    }

    public void testGetThirdPartyConnectionFromEmptyCache() {
        
        assertTrue(this.connManager.getNumberCachedFDs() == 0);
        try {
            Connection conn = this.connManager.getConnection(
                ConnectionManager.THIRDPARTY_CONNECTION, tpParams1,
                credential, TestConfig.sourceSubject,
                TestConfig.destinationSubject);
            assertTrue(this.connManager.getNumberCachedFDs() == 0);
            assertTrue(conn instanceof ThirdPartyConnectionImpl);
        } catch(Exception e) {
            fail("testGetThirdPartyConnectionFromEmptyCache() failed: " + e);
        }
    }
    
    public void testCaching() {
        
        assertTrue(this.connManager.getNumberCachedFDs() == 0);
        try {
            // get different connections
            Connection singleConn = this.connManager.getConnection(
                ConnectionManager.SINGLE_CONNECTION, singleParams,
                credential, TestConfig.sourceSubject);
            Connection mlsdConn = this.connManager.getConnection(
                ConnectionManager.MLSD_CONNECTION, mlsdParams,
                credential, TestConfig.sourceSubject);
            Connection tpConn = this.connManager.getConnection(
                ConnectionManager.THIRDPARTY_CONNECTION, tpParams1,
                credential, TestConfig.sourceSubject, 
                TestConfig.destinationSubject);
            assertTrue(this.connManager.getNumberCachedFDs() == 0);
            
            // add those connections back to the cache and check that
            // the number of cached FDs is right
            this.connManager.addConnection(singleConn);
            this.connManager.addConnection(mlsdConn);
            this.connManager.addConnection(tpConn);
            assertTrue(this.connManager.getNumberCachedFDs() ==
                (singleConn.getNumberFDs() + mlsdConn.getNumberFDs()
                    + tpConn.getNumberFDs()));
            
            // get cached connections
            Connection conn = this.connManager.getConnection(
                ConnectionManager.SINGLE_CONNECTION, singleParams,
                credential, TestConfig.sourceSubject);
            assertTrue(conn.equals(singleConn));
            conn = this.connManager.getConnection(
                ConnectionManager.MLSD_CONNECTION, mlsdParams,
                credential, TestConfig.sourceSubject);
            assertTrue(conn.equals(mlsdConn));
            conn = this.connManager.getConnection(
                ConnectionManager.THIRDPARTY_CONNECTION, tpParams1,
                credential, TestConfig.sourceSubject,
                TestConfig.destinationSubject);
            assertTrue(conn.equals(tpConn));
            
            // try to get connections with different parameters that are
            // not in the cache
            conn = this.connManager.getConnection(
                ConnectionManager.SINGLE_CONNECTION, mlsdParams,
                credential, TestConfig.sourceSubject);
            assertFalse(conn.equals(singleConn));
            conn.closeConnection();
            
            conn = this.connManager.getConnection(
                ConnectionManager.MLSD_CONNECTION, singleParams,
                credential, TestConfig.sourceSubject);
            assertFalse(conn.equals(mlsdConn));
            conn.closeConnection();
            
            conn = this.connManager.getConnection(
                ConnectionManager.THIRDPARTY_CONNECTION, tpParams2,
                credential, TestConfig.sourceSubject,
                TestConfig.destinationSubject);
            assertFalse(conn.equals(tpConn));
            conn.closeConnection();
            
        } catch(Exception e) {
            fail("testCaching() failed: " + e);
        }
    }
    
    public void testExpiration()
    {
        assertTrue(this.connManager.getNumberCachedFDs() == 0);
        try {
            Connection singleConn = this.connManager.getConnection(
                ConnectionManager.SINGLE_CONNECTION, singleParams,
                credential, TestConfig.sourceSubject);
            Connection mlsdConn = this.connManager.getConnection(
                ConnectionManager.MLSD_CONNECTION, mlsdParams,
                credential, TestConfig.sourceSubject);
            Connection tpConn = this.connManager.getConnection(
                ConnectionManager.THIRDPARTY_CONNECTION, tpParams1,
                credential, TestConfig.sourceSubject,
                TestConfig.destinationSubject);
            assertTrue(this.connManager.getNumberCachedFDs() == 0);
            
            this.connManager.addConnection(singleConn);
            this.connManager.addConnection(mlsdConn);
            this.connManager.addConnection(tpConn);
            assertTrue(this.connManager.getNumberCachedFDs() ==
                (singleConn.getNumberFDs() + mlsdConn.getNumberFDs()
                    + tpConn.getNumberFDs()));
            
            // wait longer than the connection timeout
            Thread.sleep(this.timeout+1000);            
            
            // get connections and verify that they are no longer cached
            Connection conn = this.connManager.getConnection(
                ConnectionManager.SINGLE_CONNECTION, singleParams,
                credential, TestConfig.sourceSubject);
            assertFalse(conn.equals(singleConn));
            conn.closeConnection();
            
            conn = this.connManager.getConnection(
                ConnectionManager.MLSD_CONNECTION, mlsdParams,
                credential, TestConfig.sourceSubject);
            assertFalse(conn.equals(mlsdConn));
            conn.closeConnection();
            
            conn = this.connManager.getConnection(
                ConnectionManager.THIRDPARTY_CONNECTION, tpParams1,
                credential, TestConfig.sourceSubject,
                TestConfig.destinationSubject);
            assertFalse(conn.equals(tpConn));
            conn.closeConnection();
            
            assertTrue(this.connManager.getNumberCachedFDs() == 0);
        } catch (Exception e) {
            fail("testExpiration() failed " + e);
        }
    }
}
